from bollhav.postgres.columns import PostgresColumn, PostgresType
from bollhav.postgres.schema import ensure_schema, ensure_table, ensure_schema_and_table
from bollhav.postgres.modes import (
    append,
    recreate_partition,
    recreate_table_insert,
    truncate_table_insert,
    upsert_no_delete,
    create_replace_view,
)
from bollhav.postgres.write_modes import write, write_dataframes

__all__ = [
    "PostgresColumn",
    "PostgresType",
    "ensure_schema",
    "ensure_table",
    "ensure_schema_and_table",
    "append",
    "recreate_partition",
    "recreate_table_insert",
    "truncate_table_insert",
    "upsert_no_delete",
    "create_replace_view",
    "write",
    "write_dataframes",
]



from dataclasses import dataclass
from enum import Enum


class Database(Enum):
    POSTGRES = "POSTGRES"
    PARQUET = "PARQUET"


@dataclass
class DatabaseColumn:
    name: str
    nullable: bool = True
    order: int | None = None
    sensitive: bool = False
    description: str | None = None



from bollhav.model.database import Database, DatabaseColumn
from bollhav.model.intervals import TZInterval
from bollhav.model.matching import match_models
from bollhav.model.model import Model
from bollhav.model.model_type import ModelType
from bollhav.model.sorting import sort_columns
from bollhav.model.write_modes import WriteMode
from bollhav.model.schema import Schema
from bollhav.model.source import Source
from bollhav.model.target import Target
from bollhav.model.bounds import Bounds
from bollhav.model.batch import Batch
from bollhav.model.tags import Tags
from bollhav.model.progress_bar import progress_bar

__all__ = [
    "WriteMode",
    "Database",
    "DatabaseColumn",
    "TZInterval",
    "match_models",
    "Model",
    "ModelType",
    "sort_columns",
    "Schema",
    "Source",
    "Target",
    "Bounds",
    "Batch",
    "Tags",
    "progress_bar",
]



from enum import Enum


class WriteMode(Enum):
    APPEND = "APPEND"  # Simply append rows each time
    RECREATE_TABLE_INSERT = (
        "RECREATE_TABLE_INSERT"  # Drop and create table before fully loading it
    )
    TRUNCATE_TABLE_INSERT = "TRUNCATE_TABLE_INSERT"  # Fully load a table each time
    UPSERT_NO_DELETE = "UPSERT_NO_DELETE"  # Update and/or insert, but do not delete
    RECREATE_PARTITION = (
        "RECREATE_PARTITION"  # Update and/or insert, but _do_ delete first
    )
    VIEW = "VIEW"  # Simply update view in place



import re
from dataclasses import dataclass


@dataclass
class PotentialTagMatch:
    candidates: list[str]
    reload: bool


@dataclass
class PotentialTagGroup:
    tags: list[PotentialTagMatch]


def _parse_candidates(part: str) -> list[str]:
    or_match = re.fullmatch(r"\(([^)]+)\)", part)
    if or_match:
        return or_match.group(1).split("|")
    if "|" in part:
        return part.split("|")
    return [part]


def _parse_potential_match(part: str, group_reload: bool) -> PotentialTagMatch:
    part = part.strip()
    reload = group_reload or part.startswith("r:")
    if part.startswith("r:"):
        part = part[2:]
    return PotentialTagMatch(candidates=_parse_candidates(part), reload=reload)


def _parse_group(bang: str, group_content: str) -> PotentialTagGroup:
    group_reload = bang == "r:"
    tags = [
        _parse_potential_match(part, group_reload) for part in group_content.split("&")
    ]
    return PotentialTagGroup(tags=tags)


def parse_expression(expr: str) -> list[PotentialTagGroup]:
    groups = re.findall(r"(r:)?\[([^\]]+)\]", expr)
    if not groups:
        raise ValueError(f"Invalid tag expression: {expr!r}. Must use [group] syntax.")
    return [_parse_group(bang, content) for bang, content in groups]


def group_matches(model_tags: set[str], tags: list[PotentialTagMatch]) -> bool:
    return all(any(opt in model_tags for opt in tag.candidates) for tag in tags)


def tags_match(model_tags: set[str], parsed: list[PotentialTagGroup]) -> bool:
    return any(group_matches(model_tags, group.tags) for group in parsed)



from datetime import datetime, timezone
from unittest.mock import MagicMock
import pytest
from bollhav.model.model_type import ModelType
from bollhav.model.model import Model
from bollhav.model.schema import Schema
from bollhav.model.target import Target
from bollhav.model.bounds import Bounds
from bollhav.model.tags import Tags
from bollhav.model.write_modes import WriteMode


UTC = timezone.utc


def make_column(name: str, unique: bool = False, sensitive: bool = False) -> MagicMock:
    col = MagicMock()
    col.name = name
    col.unique = unique
    col.sensitive = sensitive
    return col


def make_db() -> MagicMock:
    return MagicMock()


def make_model(**overrides) -> Model:
    return Model(
        name=overrides.pop("name", "test_model"),
        target=overrides.pop("target", Target(name="test_table")),
        source=overrides.pop("source", None),
        **overrides,
    )


# --- Target validation ---


def test_view_model_type_requires_view_write_mode():
    with pytest.raises(ValueError, match="ModelType.VIEW must use WriteMode.VIEW"):
        Target(
            name="test_table", model_type=ModelType.VIEW, write_mode=WriteMode.APPEND
        )


def test_table_model_type_cannot_use_view_write_mode():
    with pytest.raises(ValueError, match="ModelType.TABLE cannot use WriteMode.VIEW"):
        Target(name="test_table", model_type=ModelType.TABLE, write_mode=WriteMode.VIEW)


def test_database_without_columns_raises():
    with pytest.raises(ValueError, match="columns must be set"):
        Target(name="test_table", database=make_db())


def test_columns_without_database_raises():
    with pytest.raises(ValueError, match="database must be set"):
        Target(name="test_table", columns=[make_column("id")])


def test_partitioned_by_unknown_column_raises():
    with pytest.raises(ValueError, match="unknown column"):
        Target(
            name="test_table",
            database=make_db(),
            columns=[make_column("id"), make_column("ts")],
            partitioned_by="missing",
        )


def test_upsert_no_delete_without_unique_columns_raises():
    with pytest.raises(
        ValueError, match="requires at least one column with unique=True"
    ):
        Target(
            name="test_table",
            database=make_db(),
            columns=[make_column("id"), make_column("ts")],
            write_mode=WriteMode.UPSERT_NO_DELETE,
        )


# --- Bounds ---


def test_bounds_stores_begin_end():
    begin = datetime(2024, 1, 1, tzinfo=UTC)
    end = datetime(2024, 12, 31, tzinfo=UTC)
    b = Bounds(begin=begin, end=end)
    assert b.begin == begin
    assert b.end == end


# --- Model defaults ---


def test_defaults():
    m = make_model()
    assert m.source is None
    assert m.target.name == "test_table"
    assert m.target.schema.resolved == ""
    assert m.target.database is None
    assert m.target.columns == []
    assert m.target.model_type == ModelType.TABLE
    assert m.target.write_mode == WriteMode.APPEND
    assert m.enabled is True
    assert m.debug is False
    assert m.batching.default == "@daily"
    assert m.target.sensitive is False
    assert m.target.unique_columns == []
    assert m.batching.retries is None
    assert m.batching.lookback is None


# --- Tags ---


def test_name_added_to_tags_by_default():
    m = make_model()
    assert "test_model" in m.tags


def test_schema_added_to_tags_by_default():
    m = make_model(target=Target(name="test_table", schema=Schema(name="my_schema")))
    assert "my_schema" in m.tags


def test_all_tag_added_by_default():
    m = make_model()
    assert "all" in m.tags


def test_name_not_added_to_tags_when_disabled():
    m = make_model(tagging=Tags(name_add_to_tags=False))
    assert "test_model" not in m.tags


def test_schema_not_added_to_tags_when_disabled():
    m = make_model(
        target=Target(name="test_table", schema=Schema(name="my_schema")),
        tagging=Tags(schema_add_to_tags=False),
    )
    assert "my_schema" not in m.tags


def test_all_tag_not_added_when_disabled():
    m = make_model(tagging=Tags(model_gets_all_tag=False))
    assert "all" not in m.tags


# --- Target columns ---


def test_sensitive_flag_set_from_columns():
    t = Target(
        name="test_table",
        database=make_db(),
        columns=[make_column("id"), make_column("secret", sensitive=True)],
    )
    assert t.sensitive is True


def test_sensitive_false_when_no_sensitive_columns():
    t = Target(
        name="test_table",
        database=make_db(),
        columns=[make_column("id"), make_column("name")],
    )
    assert t.sensitive is False


def test_unique_columns_extracted():
    t = Target(
        name="test_table",
        database=make_db(),
        columns=[make_column("id", unique=True), make_column("name")],
    )
    assert len(t.unique_columns) == 1
    assert t.unique_columns[0].name == "id"


def test_upsert_no_delete_with_unique_column_ok():
    t = Target(
        name="test_table",
        database=make_db(),
        columns=[make_column("id", unique=True), make_column("name")],
        write_mode=WriteMode.UPSERT_NO_DELETE,
    )
    assert t.write_mode == WriteMode.UPSERT_NO_DELETE


# --- Column sorting ---


def test_column_sorting_applied():
    def custom_sort(names: list[str]) -> list[str]:
        return sorted(names, reverse=True)

    t = Target(
        name="test_table",
        database=make_db(),
        columns=[make_column("b"), make_column("a")],
        column_sorting=custom_sort,
    )
    assert [c.name for c in t.columns] == ["b", "a"]


def test_column_sorting_none_skips_sort():
    t = Target(
        name="test_table",
        database=make_db(),
        columns=[make_column("b"), make_column("a")],
        column_sorting=None,
    )
    assert [c.name for c in t.columns] == ["b", "a"]


# --- Partitioned by ---


def test_partitioned_by_sets_index_true():
    t = Target(
        name="test_table",
        database=make_db(),
        columns=[make_column("id"), make_column("ts")],
        partitioned_by="ts",
        column_sorting=None,
    )
    assert t.partitioned_by == "ts"
    assert t.partitioned_by_index is True


def test_partitioned_by_none_sets_index_false():
    t = Target(name="test_table")
    assert t.partitioned_by_index is False


# --- Extra kwargs ---


def test_extra_kwargs_stored():
    m = make_model(custom_key="custom_value")
    assert m.extra["custom_key"] == "custom_value"


def test_callable_extra_kwarg_resolved():
    m = make_model(
        static_val="hello",
        dynamic_val=lambda static_val: static_val + "_world",
    )
    assert m.extra["dynamic_val"] == "hello_world"


# --- __repr__ ---


def test_repr_contains_name():
    m = make_model()
    assert "test_model" in repr(m)


# --- __eq__ ---


def test_eq_same_config():
    m1 = make_model()
    m2 = make_model()
    assert m1 == m2


def test_eq_different_config():
    m1 = make_model(name="a")
    m2 = make_model(name="b")
    assert m1 != m2


def test_eq_non_model():
    m = make_model()
    assert m.__eq__("not a model") == NotImplemented



from bollhav.model.model import Model
from bollhav.model.target import Target


def make_model(**overrides) -> Model:
    return Model(
        name=overrides.pop("name", "orders"),
        target=overrides.pop("target", Target(name="orders")),
        source=overrides.pop("source", None),
        **overrides,
    )


def test_model_stores_fields():
    m = make_model()
    assert m.name == "orders"
    assert m.source is None


def test_model_exposes_sub_configs():
    from bollhav.model.schema import Schema
    from bollhav.model.target import Target
    from bollhav.model.batch import Batch
    from bollhav.model.bounds import Bounds

    m = make_model()
    assert isinstance(m.target.schema, Schema)
    assert isinstance(m.target, Target)
    assert isinstance(m.batching, Batch)
    assert isinstance(m.bounds, Bounds)
    assert isinstance(m.tags, set)



from icron import croniter
from datetime import datetime, timezone, timedelta


def _resolve_cron_interval(expression: str) -> tuple[datetime, datetime]:
    now = datetime.now(tz=timezone.utc)
    cron = croniter(expression, now - timedelta(days=2))
    ticks = []
    while True:
        tick = cron.get_next(datetime)
        if tick >= now:
            break
        ticks.append(tick)
    since = ticks[-2]
    until = ticks[-1]
    return since, until



from datetime import datetime, timezone
from unittest.mock import MagicMock, patch
import pytest
import sys

roskarl_mock = MagicMock()
cron_mock = MagicMock()
sys.modules["roskarl"] = roskarl_mock
sys.modules["icron"] = MagicMock()
sys.modules["cron"] = cron_mock

from bollhav.pipe.pipe_config import (  # noqa: E402
    PipeConfig,
    LatestConfig,
    BackfillConfig,
    load_pipe_config,
    with_pipe_config,
)

DT_SINCE = datetime(2024, 1, 1, tzinfo=timezone.utc)
DT_UNTIL = datetime(2024, 1, 2, tzinfo=timezone.utc)


def _base_latest(enabled: bool = False) -> LatestConfig:
    return LatestConfig(enabled=enabled, batch_expression=None, since=None, until=None)


def _base_backfill(enabled: bool = False) -> BackfillConfig:
    return BackfillConfig(
        enabled=enabled, since=None, until=None, batch_expression=None
    )


def _make_patches(
    *,
    latest_enabled: bool = False,
    backfill_enabled: bool = False,
    tags: str = "mytag",
    schema_suffix: str = "dev",
    use_schema_suffix: bool = True,
    debug: bool = False,
    cron_expr: str | None = None,
    backfill_since: datetime | None = None,
    backfill_until: datetime | None = None,
    backfill_cron: str | None = None,
    cron_interval: tuple[datetime, datetime] = (DT_SINCE, DT_UNTIL),
) -> dict:
    return {
        "bollhav.pipe.pipe_config.env_var_bool": lambda name, default=False: {
            "LATEST_ENABLED": latest_enabled,
            "BACKFILL_ENABLED": backfill_enabled,
            "DEBUG": debug,
            "USE_SCHEMA_SUFFIX": use_schema_suffix,
        }.get(name, default),
        "bollhav.pipe.pipe_config.env_var": lambda name, required=False, default=None: {
            "TAGS": tags,
            "SCHEMA_SUFFIX": schema_suffix,
        }.get(name, default),
        "bollhav.pipe.pipe_config.env_var_batch_expression": lambda name: {
            "LATEST_BATCH_EXPRESSION": cron_expr,
            "BACKFILL_BATCH_EXPRESSION": backfill_cron,
        }.get(name),
        "bollhav.pipe.pipe_config.env_var_iso8601_datetime": lambda name: {
            "BACKFILL_SINCE": backfill_since,
            "BACKFILL_UNTIL": backfill_until,
        }.get(name),
        "bollhav.pipe.pipe_config._resolve_cron_interval": MagicMock(
            return_value=cron_interval
        ),
    }


# ... rest of TestPipeConfigPostInit and TestPipeConfigDebugprint unchanged ...


class TestLoadPipeConfig:
    def _run(self, **kwargs) -> PipeConfig:
        patches = _make_patches(**kwargs)
        with (
            patch(
                "bollhav.pipe.pipe_config.env_var_bool",
                patches["bollhav.pipe.pipe_config.env_var_bool"],
            ),
            patch(
                "bollhav.pipe.pipe_config.env_var",
                patches["bollhav.pipe.pipe_config.env_var"],
            ),
            patch(
                "bollhav.pipe.pipe_config.env_var_batch_expression",
                patches["bollhav.pipe.pipe_config.env_var_batch_expression"],
            ),
            patch(
                "bollhav.pipe.pipe_config.env_var_iso8601_datetime",
                patches["bollhav.pipe.pipe_config.env_var_iso8601_datetime"],
            ),
            patch(
                "bollhav.pipe.pipe_config._resolve_cron_interval",
                patches["bollhav.pipe.pipe_config._resolve_cron_interval"],
            ),
        ):
            return load_pipe_config()

    def test_basic_load(self) -> None:
        cfg = self._run()
        assert cfg.tags == "mytag"
        assert cfg.schema_suffix == "dev"
        assert cfg.latest.enabled is False
        assert cfg.backfill.enabled is False

    def test_raises_when_both_enabled(self) -> None:
        with pytest.raises(
            ValueError, match="LATEST_ENABLED and BACKFILL_ENABLED cannot both be true"
        ):
            self._run(latest_enabled=True, backfill_enabled=True)

    def test_latest_resolves_cron_interval(self) -> None:
        cfg = self._run(latest_enabled=True, cron_expr="0 0 * * *")
        assert cfg.latest.enabled is True
        assert cfg.latest.since == DT_SINCE
        assert cfg.latest.until == DT_UNTIL
        assert cfg.latest.batch_expression == "0 0 * * *"

    def test_latest_disabled_clears_cron(self) -> None:
        cfg = self._run(latest_enabled=False, cron_expr="0 0 * * *")
        assert cfg.latest.batch_expression is None
        assert cfg.latest.since is None
        assert cfg.latest.until is None

    def test_backfill_populates_fields(self) -> None:
        cfg = self._run(
            backfill_enabled=True,
            backfill_since=DT_SINCE,
            backfill_until=DT_UNTIL,
            backfill_cron="0 6 * * *",
        )
        assert cfg.backfill.enabled is True
        assert cfg.backfill.since == DT_SINCE
        assert cfg.backfill.until == DT_UNTIL
        assert cfg.backfill.batch_expression == "0 6 * * *"

    def test_backfill_disabled_clears_fields(self) -> None:
        cfg = self._run(backfill_enabled=False)
        assert cfg.backfill.since is None
        assert cfg.backfill.until is None
        assert cfg.backfill.batch_expression is None

    def test_use_schema_suffix_false_clears_suffix(self) -> None:
        cfg = self._run(use_schema_suffix=False)
        assert cfg.schema_suffix == ""


class TestWithPipeConfig:
    def test_calls_func_with_config(self) -> None:
        patches = _make_patches(tags="t")
        received: list[PipeConfig] = []

        @with_pipe_config
        def my_func(cfg: PipeConfig) -> None:
            received.append(cfg)

        with (
            patch(
                "bollhav.pipe.pipe_config.env_var_bool",
                patches["bollhav.pipe.pipe_config.env_var_bool"],
            ),
            patch(
                "bollhav.pipe.pipe_config.env_var",
                patches["bollhav.pipe.pipe_config.env_var"],
            ),
            patch(
                "bollhav.pipe.pipe_config.env_var_batch_expression",
                patches["bollhav.pipe.pipe_config.env_var_batch_expression"],
            ),
            patch(
                "bollhav.pipe.pipe_config.env_var_iso8601_datetime",
                patches["bollhav.pipe.pipe_config.env_var_iso8601_datetime"],
            ),
            patch(
                "bollhav.pipe.pipe_config._resolve_cron_interval",
                patches["bollhav.pipe.pipe_config._resolve_cron_interval"],
            ),
        ):
            my_func()

        assert len(received) == 1
        assert received[0].tags == "t"
